/*
 * multican 0.0.5, USB remote control tool for Canon cameras.
 *
 * Copyright (C) 2006 Jindrich Novy (jnovy@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <endian.h>
#include <time.h>
#include <sched.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <usb.h>

#include "usbio.h"
#include "canonio.h"

/*
 * Purpose:
 * - scan USB ports for attached Canon devices
 *
 * Returns:
 * - number of successfully recognized Canon devices attached to USB ports
 * - filled camera[6] field
 */
int canon_scan( camera **c ) {
	struct usb_bus *busses;
	struct usb_bus *bus;
	struct usb_interface_descriptor *desc;
	int count = 0, n, ready = 0;
	camera *cam;

	FUNCSTART

	usb_init();
	usb_find_busses();
	usb_find_devices();

	busses = usb_get_busses();

	*c = calloc(USB_MAX_DEVICES, sizeof(camera));
	cam = *c;

	for (bus = busses; bus; bus = bus->next) {
                struct usb_device *dev;
		int i;

		for (dev = bus->devices; dev; dev = dev->next) {
			if (dev->descriptor.idVendor || dev->descriptor.idProduct) {
#ifdef DEBUG
				printf("Found: %04x:%04x\n", dev->descriptor.idVendor, dev->descriptor.idProduct);
#endif
				for ( i=0; camera_id[i].idVendor; i++ )
					if ( dev->descriptor.idVendor == camera_id[i].idVendor && dev->descriptor.idProduct == camera_id[i].idProduct ) {
#ifdef DEBUG
						printf(" -> %s.\n", camera_id[i].name);
#endif
						cam[count].dev = dev;
						cam[count].cam_id = i;
						
						cam[count].h = usb_open(dev);
						if ( !cam[count].h ) {
							fprintf(stderr, "  usb_open(): Unable to open USB device.\n");
							count++;
							break;
						}

						if ( usb_claim_interface(cam[count].h, 0) < 0 ) {
							fprintf(stderr, "  usb_claim_interface(): Cannot claim USB interface.\n");
							count++;
							break;
						}
						
						if ( usb_reset(cam[count].h) ) {
							fprintf(stderr, "  usb_reset(): Unable to reset interface.\n");
							count++;
							break;
						}
						
						cam[count].interface = cam[count].dev->config->interface;
						if (cam[count].interface->num_altsetting != 1) {
							fprintf(stderr,"  bad altsetting (%d).\n", cam[count].interface->num_altsetting);
							count++;
							break;
						}
						
						cam[count].desc = &cam[count].interface->altsetting[0];
						desc = cam[count].desc;

						for (n = 0; n<desc->bNumEndpoints; n++) {
							struct usb_endpoint_descriptor *endpoint = &desc->endpoint[n];

							switch(endpoint->bmAttributes & USB_ENDPOINT_TYPE_MASK) {
								case USB_ENDPOINT_TYPE_BULK:
									if ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_ENDPOINT_IN) {
										cam[count].bulk_in = endpoint->bEndpointAddress;
									} else {
										cam[count].bulk_out = endpoint->bEndpointAddress;
									}
								break;
								case USB_ENDPOINT_TYPE_INTERRUPT:
									if ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_ENDPOINT_IN) {
										cam[count].bulk_int_in = endpoint->bEndpointAddress;
									}
								break;
							}
        					}
#ifdef DEBUG
						fprintf(stderr, "  bulk_in=%d (0x%x), bulk_out=%d (0x%x), bulk_int_in=%d (0x%x)\n\n",
						cam[count].bulk_in, cam[count].bulk_in, cam[count].bulk_out, cam[count].bulk_out, cam[count].bulk_int_in, cam[count].bulk_int_in);
#endif
						if (cam[count].bulk_in == 0 || cam[count].bulk_out == 0 || cam[count].bulk_int_in == 0) {
							fprintf(stderr, "  bad endpoints.\n");
							count++;
							break;
						}

						cam[count].ready = 1;
						
						count++;
						ready++;
						if ( count>=7 ) {
							fprintf(stderr, "  maximal number of supported devices reached.\n");
							return ready;
						}
						break;
					}
                        }
                }
        }
	
	if (!ready) {
		free(cam);
		*c = NULL;
		return 0;
	}

	return ready;
}

/*
 * Purpose:
 * - initialize detected camera for communication
 *
 * Returns:
 * - 1 == ok
 * - 0 == error
 * - sets camera ready flag to 0 when initialization is unsuccessful
 * - sets max_transfer
 */
int canon_init_camera( camera *c ) {
	char buf[0x58];
	char canon_state;

	FUNCSTART
	
	if ( !c->ready ) return 0;
#ifdef DEBUG
	fprintf(stderr, "Initializing %s...\n", camera_id[c->cam_id].name);
#endif

	usb_read_byte(c->h, 0x55, &canon_state);
	c->canon_state = canon_state;

	switch( canon_state ) {
		case 'A':
#ifdef DEBUG
			printf("Camera was already active.\n\n");
#endif
			usb_read_bytes(c->h, 1, buf, sizeof(buf));
			usb_read_bytes(c->h, 4, buf, 0x50); /* XXX: requesttype 0x4 instead of 0x40? */

/*			if ( CANON_CLASS(c) >= 6 ) {
				char tmp[1024];
				usb_send_packet(c, 0x24, 0x12, 0x201, NULL, 0);
				usb_receive_packet(c, tmp, sizeof(tmp));
			}
*/
			break;
		case 'C':
#ifdef DEBUG
			printf("Camera was woken up.\n\n");
#endif
			usb_read_bytes(c->h, 1, buf, sizeof(buf));
#ifdef DEBUG
			fprintf(stderr,"Max download transfer length: %u (0x%x).\n\n", get_le32(buf+0x4c), get_le32(buf+0x4c));
#endif
			c->max_transfer = get_le32(buf+0x4c);
			buf[0] = 0x10;
			memmove(&buf[0x40], &buf[sizeof(buf)-0x10], 0x10);
			usb_write_bytes(c->h, 0x11, buf, 0x50);
			if ( CANON_CLASS(c) < 6 ) {
				usb_bulk_read_bytes(c->h, c->bulk_in, buf, 0x40);
#ifdef DEBUG
				fprintf(stderr, "Length should be 4, length read from packet is: %u (0x%x).\n\n", get_le32(buf), get_le32(buf));
#endif
				usb_bulk_read_bytes(c->h, c->bulk_in, buf, 0x4);
				usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);
			} else {
				if ( CANON_CLASS(c) >= 6 && !IS_TYPE(c,"1D")) {
#ifdef DEBUG
					fprintf(stderr, "Class >= 6 camera reading all 0x44 bytes.\n\n");
#endif
					usb_bulk_read_bytes(c->h, c->bulk_in, buf, 0x44);
				} else {
					usb_bulk_read_bytes(c->h, c->bulk_in, buf, 0x40);
					usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);
				}
			}
			break;
		case 'I':
		case 'E':
#ifdef DEBUG
			fprintf(stderr, "Camera reports an error.\n");
#endif
			c->ready = 0;
			return 0;
			break;
		default:
			fprintf(stderr, "Camera returned unknown state '%c' (0x%02x).\n", canon_state, canon_state);
			c->ready = 0;
			return 0;
	}
	
	return 1;
}

/*
 * Purpose:
 * - initialize all detected cameras.
 *
 * Returns:
 * - number of cameras successfuly initialized
 */
int canon_init_all( camera *c ) {
	int i, ready = 0;

	FUNCSTART

	for ( i=0; i<USB_MAX_DEVICES; i++ )
		if ( c[i].ready ) {
			canon_init_camera( &c[i] );
			if ( c[i].ready ) ready++;
		}

	return ready;
}

void canon_identify_camera( camera *c ) {
	char buf[0x9c];

	FUNCSTART

	if ( !c->identified )
		c->identified = 1;
	else
		return;

	usb_send_packet(c, 0x1, 0x12, 0x201, NULL, 0);
	usb_receive_packet(c, buf, sizeof(buf));

	printf("Camera type:\t\t'%s'.\nOwner:\t\t\t'%s'.\nFirmware version:\t%c.%c.%c.%c.\n",
		buf+0x5c, buf+0x7c, buf[0x5b]+'0', buf[0x5a]+'0', buf[0x59]+'0', buf[0x58]+'0');

	{
		char firmware[] = {"0.0.0.0"};
	
		c->camera_type = strdup(buf+0x5c);
		c->owner = strdup(buf+0x7c);

		firmware[0] = buf[0x5b]+'0';
		firmware[2] = buf[0x5a]+'0';
		firmware[4] = buf[0x59]+'0';
		firmware[6] = buf[0x58]+'0';
		c->firmware = strdup(firmware);
		
		c->firmware1 = buf[0x5b];
		c->firmware2 = buf[0x5a];
		c->firmware3 = buf[0x59];
		c->firmware4 = buf[0x59];
	}
}

void canon_power_status( camera *c ) {
	char buf[0x58];

	FUNCSTART
	
	if ( IS_TYPE(c,"10D") ) {
		c->power_status = strdup("unknown");
		c->power_type   = strdup("unknown");
		printf("Power status:\t\t%s.\nPower type:\t\t%s.\n", 
			c->power_status, c->power_type);
		return;
	}
	
	if ( IS_TYPE(c,"300D") || IS_TYPE(c,"10D") || IS_TYPE(c,"1D") ) {
		usb_send_packet(c, 0xa, 0x12, 0x201, NULL, 0);
	} else {
		usb_send_packet(c, 0x13, 0x12, 0x201, NULL, 0);
	}
	usb_receive_packet(c, buf, sizeof(buf));
	
	c->power_status = strdup(buf[0x54]==6?"good":buf[0x54]==4?"bad":"unknown");
	c->power_type   = strdup(buf[0x57]&0x20?"battery":"AC adapter");
	
	printf("Power status:\t\t%s.\nPower type:\t\t%s.\n", 
		c->power_status, c->power_type);

	if ( CANON_CLASS(c) == 6 )
		usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);
}

void canon_eos_id( camera *c ) {
	char buf[0x58];

	FUNCSTART
	
	if ( !c->got_body_id )
		c->got_body_id = 1;
	else
		return;
	
	if ( !IS_EOS(c) ) {
#ifdef DEBUG
		fprintf(stderr, "Unable to get body ID on non-EOS cameras.\n");
#endif
		exit(1);
	}
	
	if ( CANON_CLASS(c) >= 6 /*&& !IS_TYPE(c,"1D")*/ ) {
		usb_send_packet(c, 0x23, 0x12, 0x201, NULL, 0);
	} else {
		usb_send_packet(c, 0x1d, 0x12, 0x201, NULL, 0);
	}
	usb_receive_packet(c, buf, sizeof(buf));
	
	c->body_id = get_le32(&buf[0x54]);
	
	printf("Serial number:\t\t%010u.\n", c->body_id);
}

void canon_init_control( camera *c ) {
	char buf[0x5c];
	char args[0x9];

	FUNCSTART

	if ( IS_TYPE(c,"300D") || IS_TYPE(c,"10D") || IS_TYPE(c,"1D") ) {
		memset(args, 0, 8);
		usb_send_packet( c, 0x13, 0x12, 0x201, args, 8 );
		usb_receive_packet(c, buf, sizeof(buf));
	} else {
		memset(args, 0, 9);
		usb_send_packet( c, 0x25, 0x12, 0x201, args, 9 );
		usb_receive_packet(c, buf, sizeof(buf));
	}
}

/* Doesn't work with EOS 300D */
#define FLASH_OFF  0x00
#define FLASH_ON   0x01
#define FLASH_AUTO 0x02
void canon_flash( camera *c, unsigned char flags ) {
	char buf[0x5c];
	char args[0x38];

	FUNCSTART
	
	printf("Flash is set:\t\t");
	switch( flags ) {
		case FLASH_OFF:
			printf("OFF\n");
		break;
		case FLASH_ON:
			printf("ON\n");
		break;
		case FLASH_AUTO:
			printf("AUTO\n");
		break;
		default:
			printf("to unknown state, THIS IS A BUG!\n");
			return;
		break;
	}
	memset(args, 0, sizeof(args));
	args[0]    = 7;
	args[0x12] = flags;
	
	c->flash_state = flags;

	usb_send_packet( c, 0x13, 0x12, 0x201, args, sizeof(args) );
	usb_receive_packet(c, buf, sizeof(buf));
}

void canon_transfer( camera *c, unsigned char flags ) {
	char buf[0x5c];
	char args[0xc];
	
	FUNCSTART
	
	if ( flags&~15 ) {
		fprintf(stderr, "canon_transfer(): Flags not in range!\n");
	}
	
	c->thumb_to_pc = flags&THUMB_TO_PC;
	c->full_to_pc = flags&FULL_TO_PC;
	c->thumb_to_drive = flags&THUMB_TO_DRIVE;
	c->full_to_drive = flags&FULL_TO_DRIVE;
	
	printf("Transfer mode is:%s%s%s%s.\n",
		c->thumb_to_pc?"\tTHUMB_TO_PC":"",
		c->full_to_pc?"\tFULL_TO_PC":"",
		c->thumb_to_drive?"\tTHUMB_TO_DRIVE":"",
		c->full_to_drive?"\tFULL_TO_DRIVE":""
	      );

	memset(args, 0, sizeof(args));

	args[0] = 9;
	args[4] = 4;
	args[8] = flags;

	if ( IS_TYPE(c,"300D") || IS_TYPE(c,"10D") )
		usb_send_packet( c, 0x13, 0x12, 0x201, args, sizeof(args) );
	else
		usb_send_packet( c, 0x25, 0x12, 0x201, args, sizeof(args) );

	usb_receive_packet(c, buf, sizeof(buf));
}

/* Doesn't work on EOS 300D */
void canon_get_zoom( camera *c ) {
	char buf[0x60];
	char args[0x8];

	FUNCSTART

	args[0] = 0xb;
	memset(args+1, 0, 7);
	usb_send_packet( c, 0x13, 0x12, 0x201, args, sizeof(args) );
	usb_receive_packet(c, buf, sizeof(buf));
	
	printf("Zoom:\t\t%d\n", buf[0x5e]);
}

void canon_avail_shots( camera *c ) {
	char buf[0x60];
	char args[0x8];
	
	FUNCSTART

	args[0] = 0xd;
	memset(args+1, 0, 7);
	usb_send_packet( c, 0x13, 0x12, 0x201, args, sizeof(args) );
	usb_receive_packet(c, buf, sizeof(buf));
	
	c->avail_shots = get_le32(&buf[0x5c]);
	printf("Available shots:\t%u\n", c->avail_shots);
}

void canon_exit_control( camera *c ) {
	char buf[0x5c];
	char args[0x8];

	FUNCSTART

	args[0] = 1;
	memset(args+1, 0, 7);

	if ( IS_TYPE(c,"300D")|| IS_TYPE(c,"10D") )
		usb_send_packet( c, 0x13, 0x12, 0x201, args, sizeof(args) );
	else
		usb_send_packet( c, 0x25, 0x12, 0x201, args, sizeof(args) );

	usb_receive_packet(c, buf, sizeof(buf));
}

void canon_close( camera **cptr ) {
	int i;

	FUNCSTART
	
	if ( !*cptr ) return;
	
	for ( i=0; i<USB_MAX_DEVICES; i++ ) {
		camera *c = &(*cptr)[i];
	
		free(c->owner);
		free(c->camera_type);
		free(c->firmware);
		free(c->power_status);
		free(c->power_type);
		
		if ( c->interface ) usb_release_interface(c->h, 0);
		if ( c->h ) usb_close(c->h);
	}
	
	free( *cptr );
	*cptr = NULL;
}

void canon_set_owner( camera *c, char *name ) {
	usb_send_packet( c, 0x1, 0x12, 0x201, name, strlen(name)+1 );
}

/*
 * Has to be called before changing any of the release parameters.
 */
void canon_get_rel_params( camera *c ) {
	char buf[0x8c];
	char args[0x8];

	FUNCSTART
	
	memset(args, 0, sizeof(args));
	args[0]    = 0x0a;

	if ( IS_TYPE(c,"300D") || IS_TYPE(c,"10D") )
		usb_send_packet( c, 0x13, 0x12, 0x201, args, sizeof(args) );
	else
		usb_send_packet( c, 0x25, 0x12, 0x201, args, sizeof(args) );

	usb_receive_packet(c, buf, sizeof(buf));
	
	memcpy(c->rel_params, &buf[0x5c], 0x30);
	c->rel_params_filled = 1;
	c->rel_params_updated = 0;
}

void canon_set_rel_params( camera *c ) {
	char buf2[0x5c];
	char args2[0x38];

	FUNCSTART
	
	if (!c->rel_params_filled) {
		fprintf(stderr, "warning: Release parameters not previously read.\n");
		return;
	}

	memset(args2, 0, sizeof(args2));
	args2[0] = 7;
	args2[4] = 0x30;
	memcpy(&args2[8], c->rel_params, 0x30);

	if ( IS_TYPE(c,"300D") || IS_TYPE(c,"10D") )
		usb_send_packet( c, 0x13, 0x12, 0x201, args2, sizeof(args2) );
	else
		usb_send_packet( c, 0x25, 0x12, 0x201, args2, sizeof(args2) );

	usb_receive_packet(c, buf2, sizeof(buf2));

	c->rel_params_updated = 0;
}

void canon_rel( camera *c, char *param, char *token ) {
	int par_no, i, j;

	FUNCSTART
	
	/* first, find the correct struct */
	for ( i=0; rel_param_entries[i].name; i++ ) {
		if ( !strcmp(rel_param_entries[i].name, param) ) {
#ifdef DEBUG
			fprintf(stderr, "canon_rel: updating parameter=%s.\n", rel_param_entries[i].name);
#endif
			par_no = i;
			break;
		}
	}

	if ( !rel_param_entries[i].name ) {
		fprintf(stderr, "warning: rel parameter=%s -> no such parameter -> ignoring !!!\n", param);
		return;
	}

	/* we have a valid parameter, maybe we need to load the parameter set from the camera first */
	if (!c->rel_params_filled)
		canon_get_rel_params(c);

	/* lets see, what we should set the parameter to */
	for ( i=0; rel_param_entries[par_no].table[i].token; i++ ) {
		if ( !strcmp(rel_param_entries[par_no].table[i].token, token) ) {
#ifdef DEBUG
			fprintf(stderr, "canon_rel: setting %s=%s.\n", rel_param_entries[par_no].name, rel_param_entries[par_no].table[i].token);
#endif
			/* update entry/ies */
			for ( j=0; rel_param_entries[par_no].offset[j]; j++ ) {
				c->rel_params[(int)rel_param_entries[par_no].offset[j]] = rel_param_entries[par_no].table[i].code[j];
			}
			c->rel_params_updated = 1;
			break;
		}
	}

	if ( !rel_param_entries[par_no].table[i].token )
		fprintf(stderr,"warning: %s=%s -> no such token -> ignoring !!!\n", rel_param_entries[par_no].name, token);
}

void canon_release_shutter( camera *c ) {
	char buf[0x5c];
	char args[0x9];

	FUNCSTART

	if (c->rel_params_updated) {
		/* parameters have been changed, transmit them to camera before shooting */
		canon_set_rel_params(c);
	}

	args[0] = 4;
	memset(args+1, 0, 8);

	if ( IS_TYPE(c,"300D") || IS_TYPE(c,"10D") )
		usb_send_packet( c, 0x13, 0x12, 0x201, args, 8 );
	else
		usb_send_packet( c, 0x25, 0x12, 0x201, args, 9 );

	usb_receive_packet(c, buf, sizeof(buf));
	
/*	if ( CANON_CLASS(c) >= 6 && !IS_TYPE(c,"1D") ) {
		char buf2[0x54];
		char args2[4];

		args2[0] = 0xf;
		memset(args2+1, 0, 3);
		usb_send_packet( c, 0x22, 0x12, 0x201, args2, 4 );
		usb_receive_packet(c, buf2, sizeof(buf2));
	}
*/
	usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);

	if ( c->thumb_to_pc ) {
		fprintf(stderr, "  THUMB_TO_PC not implemented!\n");
	}

	if ( c->full_to_pc ) {
		unsigned idx, size;

		if ( buf[0xc] != 0xc ) {
			fprintf(stderr,"Warning: FULL_TO_PC interrupt read error.\n");
		}

		usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x17);

		idx  = get_le32(buf+0xc);
		size = get_le32(buf+0x11);
		printf("  size = %u B\n", size);
		
		if ( CANON_CLASS(c) != 6 ) /* 20D */
			usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);
		
		{
			char recv[0x40];
			char args[0x10];
			char *img;
			unsigned read = 0, read_chunk, actually_read;
			
			put_le32(args, 0);
/*			put_le32(&args[4], c->max_transfer); */
			put_le32(&args[4], 0x1400);
			put_le32(&args[8], 2); /* 1 - thumbnail, 2 - full image  */
			put_le32(&args[0xc], idx);

			usb_send_packet( c, 0x17, 0x12, 0x202, args, sizeof(args) );
			usb_receive_packet(c, recv, sizeof(recv));

			size = get_le32(&recv[6]);
			printf("  size = %u B\n", size);
			
			img = calloc(size, 1);
/*			usb_bulk_read_bytes(c->h, c->bulk_in, img, 0x1400);
			usb_receive_packet(c, img, size);*/

			do {
				if ( size - read > 0x1400 )
					read_chunk = 0x1400;
				else if ( size - read > 0x40 )
					read_chunk = (size - read) / 0x40 * 0x40;
				else
					read_chunk = size - read;

				actually_read = usb_bulk_read(c->h, c->bulk_in, img + read, read_chunk, USB_TIMEOUT);
/*				usb_bulk_read_bytes(c->h, c->bulk_in, img + read, read_chunk);*/

				if ( actually_read != read_chunk ) {
					fprintf(stderr, "Read error!\n");
					exit(1);
				}

				read += read_chunk;
			} while ( read < size );

			{
				FILE *f = fopen("img.jpg","wb");
				fwrite(img, size, 1, f);
				fclose(f);
			}

			free(img);
		}
	}

	if ( c->full_to_drive )	 {
		if ( buf[0xc] != 0x1c && (IS_TYPE(c,"300D") || IS_TYPE(c,"10D"))) {
			fprintf(stderr,"Warning: FULL_TO_DRIVE first interrupt read error.\n");
		}

		if ( IS_TYPE(c,"300D") || IS_TYPE(c,"10D") ) {
			usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);
			if ( buf[0xc] != 0x1d ) {
				fprintf(stderr,"Warning: FULL_TO_DRIVE second interrupt read error.\n");
			}
		}
		
		if ( IS_TYPE(c,"300D") ||  IS_TYPE(c,"10D") ) {
			usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);
			if ( buf[0x4] != 0xe ) {
				fprintf(stderr,"Warning: FULL_TO_DRIVE error saving image to CF.\n");
			}
		}
	}

	if ( IS_EOS(c) && !(IS_TYPE(c,"300D") || IS_TYPE(c,"10D")) && !IS_TYPE(c,"350D") && !IS_TYPE(c,"5D") && !IS_TYPE(c,"20D"))
		usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);

/*	if ( CANON_CLASS(c) >= 6 && !IS_TYPE(c,"1D") ) {
		char buf2[0x54];
		char args2[4];

		args2[0] = 0xf;
		memset(args2+1, 0, 3);
		usb_send_packet( c, 0x21, 0x12, 0x201, args2, 4 );
		usb_receive_packet(c, buf2, sizeof(buf2));
	}
*/
	c->rel_params_filled = 0;
}
