/*
 * multican 0.0.5, USB remote control tool for Canon cameras.
 *
 * Copyright (C) 2006 Jindrich Novy (jnovy@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <endian.h>
#include <time.h>
#include <sched.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <usb.h>

#include "usbio.h"
#include "canonio.h"

int scr_line;
char *scr, *scr_name;
off_t scr_size, scr_pos;

char *next_line() {
	while (scr_pos<scr_size && (scr[scr_pos] == '#' || scr[scr_pos] == '\n') ) {
		for (; scr_pos<scr_size && scr[scr_pos] != '\n'; scr_pos++);
		scr_pos++;
		scr_line++;
	}

	if (scr_pos<scr_size) {
		return &scr[scr_pos];
	}

	return NULL;
}

char *skip_line() {
	while (scr_pos<scr_size && scr[scr_pos] != '\n') scr_pos++;

	return next_line();
}

off_t skip_text() {
	off_t opos = scr_pos;

	while (scr_pos<scr_size && scr[scr_pos] != ' ' && scr[scr_pos] != '\t' && scr[scr_pos] != '\n') scr_pos++;

	return scr_pos-opos;
}

off_t skip_spaces() {
	off_t opos = scr_pos;

	while (scr_pos<scr_size &&
	       (scr[scr_pos] == ' ' || scr[scr_pos] == '\t') && scr[scr_pos] != '\n' && scr[scr_pos] != '\r') scr_pos++;

	return scr_pos-opos;
}

off_t scr_textsize() {
	off_t sz = skip_text();
	scr_pos -= sz;
	return sz;
}

void scr_err( char *message ) {
	fprintf(stdout, "%s:\tError at %d: %s.\n", scr_name, scr_line, message);
	exit(1);
}

void scr_warn( char *message ) {
	fprintf(stdout,"%s:\tWarning at %d: %s.\n", scr_name, scr_line, message);
}

void scr_rel( camera *c, char *token ) {
	skip_spaces();

	if ( scr[scr_pos] != '=' ) {
		scr_warn("'=' expected, skipping..");
		return;
	}

	scr_pos++;

	skip_spaces();

	{
		off_t size = scr_textsize();
		char buf[0x10];
		
		if ( size>sizeof(buf)-1 ) {
			scr_warn("value definition too long, skipping..");
			return;
		}
		
		memcpy(buf, &scr[scr_pos], size);
		buf[size] = '\0';
		canon_rel(c, token, buf);
	}
}

void scr_shot( camera *c, char *token ) {
	canon_release_shutter(c);
}

static struct {
	char *token;
	void (*cback)(camera *, char *);
} scr_cback[] = {
	{ "quality", scr_rel },
	{ "sequence", scr_rel },
	{ "beep", scr_rel },
	{ "mode", scr_rel },
	{ "auto-exp-mode", scr_rel },
	{ "white-balance", scr_rel },
	{ "auto-focus", scr_rel },
	{ "iso", scr_rel },
	{ "av", scr_rel },
	{ "tv", scr_rel },
	{ "exp-corr", scr_rel },
	{ "shot", scr_shot },
	{ "nop", NULL },
	{ NULL, NULL }
};

#ifdef ABSOLUTE_TIME
struct timeval start_tv;
#endif

void process_script( camera *c ) {
	int i;

	FUNCSTART
	
	scr_pos = 0;
	scr_line = 1;

	if ( !next_line() )
		scr_err("Unexpected EOF");

	canon_init_control(c);

	{
		int h, m, s, u;
		struct timeval tv1, tv2;
#ifndef ABSOLUTE_TIME
		struct timeval tv_tmp;
#endif

		while ( 1 ) {

			if ( sscanf(&scr[scr_pos], "%02d:%02d:%02d.%03d", &h, &m, &s, &u) != 4 ) {
				scr_warn("Invalid time format, skipping..");
				if ( !skip_line() ) break;
				continue;
			}

			if ( h<0 || m<0 || s<0 || u<0 ) {
				scr_warn("Time values cannot be negative, skipping..");
				if ( !skip_line() ) break;
				continue;
			}

			if ( h>23 || m>59 || s>59 || u>999 ) {
				scr_warn("At least on time value ot of range, skipping..");
				if ( !skip_line() ) break;
				continue;
			}

			skip_text();
			skip_spaces();

			for ( i=0; scr_cback[i].token; i++ ) {
				if (!memcmp(scr_cback[i].token, &scr[scr_pos], strlen(scr_cback[i].token))) {
					scr_pos += strlen(scr_cback[i].token);
#ifdef ABSOLUTE_TIME
					{
						int n = scr_pos;

						gettimeofday(&tv1, NULL);
						tv2.tv_sec   = start_tv.tv_sec + h*3600 + m*60 + s + ((start_tv.tv_usec + u*1000) / 1000000);
						tv2.tv_usec  = (start_tv.tv_usec + u*1000) % 1000000;

/*						printf("---\nstv: %ld:%06ld\n", start_tv.tv_sec, start_tv.tv_usec);
						printf("tv1: %ld:%06ld -> %02d:%02d:%02d.%03d\n", tv1.tv_sec, tv1.tv_usec, h, m, s, u);
						printf("tv2: %ld:%06ld\n", tv2.tv_sec, tv2.tv_usec);
*/
/*							while( nanosleep(&nano_tv, &nano_tv) == -1);
							usleep((nano_tv.tv_sec-tv1.tv_sec)*1000000);
*/						if ( tv2.tv_sec >= tv1.tv_sec ) {
							struct timespec ntv;
							if ( tv2.tv_sec == tv1.tv_sec ) {
								ntv.tv_sec  = 0;
								ntv.tv_nsec = (tv2.tv_usec - tv1.tv_usec) * 1000;

								if ( ntv.tv_nsec > 0 ) {
/*									printf("%ld:%09ld\t", ntv.tv_sec, ntv.tv_nsec); */
									while( nanosleep(&ntv, &ntv) == -1);
								}
							} else {
								long long dt = (1000000-tv1.tv_usec)+tv2.tv_usec+1000000*((tv2.tv_sec-tv1.tv_sec)-1);
								ntv.tv_sec  = dt / 1000000;
								ntv.tv_nsec = (dt % 1000000) * 1000;

/*								printf("%ld:%09ld\t", ntv.tv_sec, ntv.tv_nsec);*/
								while( nanosleep(&ntv, &ntv) == -1);
							}
						}

						fputs(scr_name,stdout);
						putchar('\t');
						printf("%02d:%02d:%02d.%03d\t", h, m, s, u);
						fputs(scr_cback[i].token, stdout);
						for ( ; n<scr_size && scr[n] != '\n'; n++ )
							putchar(scr[n]);
						putchar('\n');

						if ( scr_cback[i].cback ) scr_cback[i].cback(c, scr_cback[i].token);
					}
#else /* ABSOLUTE_TIME */
					{
						gettimeofday(&tv1, NULL);
						tv2.tv_sec  = tv1.tv_sec + h*3600 + m*60 + s + ((tv1.tv_usec + u*1000) > 1000000 ? 1 : 0);
						tv2.tv_usec = (tv1.tv_usec + u*1000) % 1000000;

						fprintf(stderr, "%s\tStart: %ld:%06ld -> %02d:%02d:%02d.%03d\n", scr_name, tv1.tv_sec, tv1.tv_usec, h, m, s, u);
						fprintf(stderr, "%s\tEnd  : %ld:%06ld\n\n", scr_name, tv2.tv_sec, tv2.tv_usec);

						if ( scr_cback[i].cback ) scr_cback[i].cback(c, scr_cback[i].token);
						
						while ( 1 ) {
							usleep(1);
							gettimeofday(&tv_tmp, NULL);
							if ( tv_tmp.tv_sec < tv2.tv_sec ) {
								continue;
							} else {
								if ( tv_tmp.tv_sec == tv2.tv_sec ) {
									if ( tv_tmp.tv_usec < tv2.tv_usec )
										continue;
								}
								break;
							}
						}
					}
#endif /* ABSOLUTE_TIME */
					break;
				}
			}
			
			if ( !scr_cback[i].token )
				scr_warn("Unknown token");

			if ( !skip_line() ) break;
		}
	}

	canon_exit_control(c);
}

int main( int argc, char **argv) {
	camera *cameras;

	printf("Multican v0.0.5, by Jindrich Novy (jnovy@users.sourceforge.net)\n\n");

	if ( !canon_scan(&cameras) ) {
		fprintf(stderr, "No known Canon camera recognized.\n");
		return 1;
	}

	if ( !canon_init_all(cameras) ) {
		fprintf(stderr, "No recognized camera was successfully initialized.\n");
		return 1;
	}

	if ( argc > 1 ) {
		int i;
		char ch;
		char *scrs[USB_MAX_DEVICES];
		char *scr_names[USB_MAX_DEVICES];
		off_t scr_sizes[USB_MAX_DEVICES];
		camera *cams[USB_MAX_DEVICES];

		printf("User request to control %d camera%s.\n\n", argc-1, argc-1>1?"s":"");
		
		if ( argc-1 > USB_MAX_DEVICES ) {
			fprintf(stderr, "More scripts than the maximum possible USB devices specified.\n");
			exit(1);
		}

		for ( i=1; i<argc ; i++ ) {
			FILE *f = fopen(argv[i],"rt");
			int n, l;
			unsigned serno;
			
			scr_name = argv[i];
			scr_names[i-1] = scr_name;

			if ( !f )
				scr_err("Unable to open script");

			l = strlen(scr_name);

			if ( l < 10 )
				scr_err("Invalid script name -> must contain serial number of a camera at the end");

			for ( n=l-10; n<l; n++ )
				if ( scr_name[n] < '0' && scr_name[n] > '9' )
					scr_err("Serial number in script name contains invalid character");

			if ( sscanf(&scr_name[l-10], "%u", &serno) != 1 )
				scr_err("Bad serial number in script name");

			fseek(f, 0, SEEK_END);
			scr_sizes[i-1] = ftell(f);
			fseek(f, 0, SEEK_SET);

			scrs[i-1] = malloc(scr_sizes[i-1]);

			if (!fread(scrs[i-1], 1, scr_sizes[i-1], f))
				scr_err("Error reading script");

			fclose(f);

			for ( n=0; n<USB_MAX_DEVICES; n++ ) {
				camera *c = &cameras[n];

				if ( !c->ready ) continue;
		
				canon_identify_camera(c);
				canon_eos_id(c);

				if ( c->body_id == serno ) {
					putchar('\n');
					cams[i-1] = c;
					break;
				}
			}

			if ( n >= USB_MAX_DEVICES )
				scr_err("No active device matches serial number in script");
		}

		printf("\nPress ENTER to start the scripted sequence.\n");
		fscanf(stdin, "%c", &ch);
		printf("Sequence started...\n");

		{
			pid_t pids[USB_MAX_DEVICES];
			int totpids = 0, status = 0;
#ifdef ABSOLUTE_TIME
			gettimeofday(&start_tv, NULL);
#endif
			for ( i=1; i<argc; i++ ) {
				if ( !(pids[totpids++]=fork()) ) {
					scr_name = scr_names[i-1];
					scr_size = scr_sizes[i-1];
					scr = scrs[i-1];

					process_script( cams[i-1] );

					return 0;
				}
			}

			for ( i=0; i<totpids; i++ ) {
				struct sched_param schp;
				
				schp.sched_priority = SCHED_RR;
			
				sched_setscheduler(pids[i], SCHED_RR, &schp);
				waitpid(pids[i], &status, WUNTRACED);

				if ( !WIFEXITED(status) ) {
					scr_name = scr_names[i];
					scr_warn("Script failed");
				}
			}
		}
	} else
	{
		int i;

		tolerate_error = 1;

		for ( i=0; i<USB_MAX_DEVICES; i++ ) {
			camera *c = &cameras[i];

			if ( !c->ready ) continue;

			canon_identify_camera(c);
			if ( !IS_TYPE(c,"10D") ) canon_power_status(c);
			if ( !IS_TYPE(c,"1D") )  canon_eos_id(c);

			canon_init_control(c);
			canon_rel(c, "quality", "fine-l");
			canon_rel(c, "tv", "0\"5");
			canon_rel(c, "av", "5.6");
			canon_rel(c, "iso", "200");

			canon_transfer(c, FULL_TO_DRIVE);

			canon_release_shutter(c);

			sleep(3);

			canon_rel(c, "quality", "coarse-m");
			canon_rel(c, "tv", "0\"4");
			canon_rel(c, "av", "6.3");
			canon_rel(c, "iso", "100");
			
			canon_release_shutter(c);

			sleep(3);

			canon_rel(c, "quality", "fine-l");
			canon_rel(c, "tv", "0\"3");
			canon_rel(c, "av", "7.1");
			canon_rel(c, "iso", "400");

			canon_release_shutter(c);

			sleep(3);
			canon_exit_control(c);
		}
	}

	canon_close(&cameras);

	return 0;
}
