/*
 * multican 0.0.5, USB remote control tool for Canon cameras.
 *
 * Copyright (C) 2006 Jindrich Novy (jnovy@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <endian.h>
#include <usb.h>

#include "usbio.h"

unsigned usb_sequence = 0;
int tolerate_error = 0, retries = 5000;

void usb_hexdump( char *bytes, int size ) {
	int n, i;
	
	fputs("         0  1  2  3  4  5  6  7| 8  9  A  B  C  D  E  F 01234567|89ABCDEF\n", stderr);
	for ( n=0; n<size; n+=i ) {
		fprintf(stderr, "0x%04X: ", n);
		for ( i=0; n+i<size && i<0x10; i++ ) {
			unsigned char b = bytes[n+i];
			fprintf(stderr, "%02X%c", b, i!=7?' ':'|');
		}
		for ( ; i<0x10; i++ )
			fputs("   ", stderr);
		for ( i=0; n+i<size && i<0x10; i++ ) {
			unsigned b = bytes[n+i];
			if ( i==8 )
				putc('|', stderr);
			putc(b>=0x20 && b<0x7f ? b : '.', stderr);
		}
		putc('\n', stderr);
	}
	putc('\n', stderr);
}

unsigned get_le32( char *b ) {
	unsigned char A, B, C, D;
#if __BYTE_ORDER == __LITTLE_ENDIAN	
	A = b[0];
	B = b[1];
	C = b[2];
	D = b[3];
#elif __BYTE_ORDER == __BIG_ENDIAN
	A = b[3];
	B = b[2];
	C = b[1];
	D = b[0];
#else
#	error "Endianess of your machine is not supported."
#endif
	return A | (B<<8) | ((C)<<16) | ((D)<<24);
}

/* Works only for little endian machines at the time! */
void put_le32( char *b, unsigned v ) {
#if __BYTE_ORDER == __LITTLE_ENDIAN
	b[0] = v&0xff;
	b[1] = (v>>8)&0xff;
	b[2] = (v>>16)&0xff;
	b[3] = (v>>24)&0xff;
#elif __BYTE_ORDER == __BIG_ENDIAN
	b[3] = v&0xff;
	b[2] = (v>>8)&0xff;
	b[1] = (v>>16)&0xff;
	b[0] = (v>>24)&0xff;
#else
#	error "Endianess of you machine is not supported."
#endif
}

void usb_read_byte( usb_dev_handle *h, int value, char *byte ) {
	int ret;
#if DEBUG
	fprintf(stderr, "usb_read_byte: value=%d (0x%x)", value, value);
#endif
	while ( (ret=USB_READ_BYTE(h, value, byte)) == 0  ) {
#if DEBUG
		fprintf(stderr, "Read nothing, trying again.\n");
#endif
	}
	
	if ( ret < 0 ) {
		fprintf(stderr, "Read failed, exiting...\n");
		exit(1);
	}
#if DEBUG
    fprintf(stderr,", byte=%c (0x%02x)\n", *(unsigned char*)byte, *(unsigned char*)byte);
#endif
}

void usb_read_bytes( usb_dev_handle *h, int value, char *bytes, int size ) {
	int ret;
#if DEBUG
	fprintf(stderr, "usb_read_bytes: value=%d (0x%x), size=%d (0x%x):\n", value, value, size, size);
#endif
	while ( (ret=USB_READ_BYTES(h, value, bytes, size)) < size  ) {
#if DEBUG
		fprintf(stderr, "Partial read, trying again.\n");
#endif
	}
	
	if ( ret < 0 ) {
		fprintf(stderr, "Read failed, exiting...\n");
		exit(1);
	}
#if DEBUG
	usb_hexdump(bytes, size);
#endif
}

void usb_write_byte( usb_dev_handle *h, int value, char *byte ) {
	int ret = 0, retr = 0;
#if DEBUG
	fprintf(stderr, "usb_write_byte: value=%d (0x%x), byte=%c (0x%02x)\n", value, value, *byte, *byte);
#endif

	while ( (ret=USB_WRITE_BYTE(h, value, byte)) == 0  ) {
#if DEBUG
		fprintf(stderr, "Wrote partially, trying again.\n");
#endif
		if ( tolerate_error && ++retr >= retries ) {
			fprintf(stderr,"Writing sequence tried %d times, giving up...\n", retr);
			return;
		}
	}

	if ( ret < 0 ) {
		fprintf(stderr, "Write failed, exiting...\n");
		exit(1);
	}
}

void usb_write_bytes( usb_dev_handle *h, int value, char *bytes, int size ) {
	int ret = 0, retr = 0;
#if DEBUG
	fprintf(stderr, "usb_write_bytes: value=%d (0x%x), size=%d (0x%x):\n", value, value, size, size);
	usb_hexdump(bytes, size);
#endif
	while ( (ret=USB_WRITE_BYTES(h, value, bytes, size)) < size  ) {
#if DEBUG
		fprintf(stderr, "Wrote partially, trying again.\n");
#endif
		if ( tolerate_error && ++retr >= retries ) {
			fprintf(stderr,"Writing sequence tried %d times, giving up...\n", retr);
			return;
		}
	}
	
	if ( ret < 0 ) {
		fprintf(stderr, "Write failed, exiting...\n");
		exit(1);
	}
}

void usb_bulk_read_bytes( usb_dev_handle *h, int ep, char *bytes, int size ) {
	int ret = -1, pos = 0, retr = 0;

	if ( !size ) {
		fprintf(stderr, "warning: bulk_read request of 0 bytes!\n");
		return;
	}
	
	memset(bytes, 0, size);
	while ( pos != size && (ret = usb_bulk_read( h, ep, (char*)&bytes[pos], size-pos, USB_TIMEOUT )) >= 0 ) {
		pos += ret;
#if DEBUG
		fprintf(stderr, "usb_bulk_read: ep=%d (0x%x), size=%d (0x%x):\n", ep, ep, ret, ret);
#endif
		retr++;
		if ( tolerate_error && retr >= retries ) break;
	}
	
	if ( pos != size || ret < 0 ) {
		fprintf(stderr, "Error in usb_bulk_read() request.\n");
		if ( !tolerate_error ) {
			exit(1);
		} else {
			fprintf(stderr, "Tolerating the above error for now.\n");
		}
	}
#if DEBUG
	usb_hexdump(bytes, size);
#endif
}

void usb_send_packet( camera *c, unsigned char cmd1, unsigned char cmd2, unsigned cmd3, char *arguments, int argsize ) {
#ifdef DEBUG
	fprintf(stderr, "usb_send_packet: cmd1=%d (0x%x), cmd2=%d (0x%x), cmd3=%d (0x%x):\n", cmd1, cmd1, cmd2, cmd2, cmd3, cmd3);
#endif
	memset(usb_packet, 0, 0x50);
	put_le32(usb_packet, 0x10+argsize);
	put_le32(usb_packet+0x48, 0x10+argsize);
	usb_packet[0x40] = 0x2;
	usb_packet[0x44] = cmd1;
	usb_packet[0x47] = cmd2;
	put_le32(usb_packet+4, cmd3);

	if ( CANON_CLASS(c) >= 6 && !IS_TYPE(c,"1D") ) {
		usb_packet[0x46] = cmd3==0x202 ? 0x20 : 0x10;
	}

	put_le32(usb_packet+0x4c, usb_sequence++);

	if (argsize < 0 || argsize+0x50 > sizeof(usb_packet)) {
		fprintf(stderr, "Invalid argsize = %d (0x%x).\n\n", argsize, argsize);
		exit(1);
	}

	if ( argsize > 0 ) {
		memcpy(usb_packet+0x50, arguments, argsize);
	}

	usb_write_bytes(c->h, 0x10, usb_packet, 0x50+argsize);
}

void usb_receive_packet( camera *c, char *bytes, int size ) {
#ifdef DEBUG
	fprintf(stderr, "usb_receive_packet: ep=%d (0x%x), size=%d (0x%x):\n", c->bulk_in, c->bulk_in, size, size);
#endif
	usb_bulk_read_bytes(c->h, c->bulk_in, bytes, size);

	{
		int i;
		unsigned code = get_le32(&bytes[0x50]);
#ifdef DEBUG
		unsigned cmd3 = ((bytes[5]-1)<<8)+bytes[4];

		if (cmd3==0x202) { /* Format of response block is different in case cmd3==0x202*/
			fprintf(stderr,"usb_receive_packet sanity check: unformatted binary data of size=%d follows\n", get_le32(&bytes[6]));
		} else {
			fprintf(stderr,"usb_receive_packet sanity check: cmd1=0x%02x, cmd2=0x%02x, cmd3=0x%03x, 2==%d, size=0x%04x\n",
			bytes[0x44], 
			bytes[0x47]-0x10,
			cmd3,
			bytes[0x40],
			get_le32(&bytes[0x48])+0x40);
		}
#endif
		if ( !code ) return;

		for ( i=0; canon_status[i].code; i++ )
			if ( canon_status[i].code == code ) {
				fprintf(stderr, "usb_receive_packet status code: 0x%08x -> %s.\n", code, canon_status[i].text);
				return;
			}
		fprintf(stderr, "usb_receive_packet exit code: 0x%08x -> UNKNOWN !!!\n", code);
	}
}
